@@ -1623,6 +1623,8 @@ enum_parse_values(
 	}
     }
 
+    p = skipwhite(p);
+
     if (*p != NUL && *p != '#')
     {
 	if (did_emsg == did_emsg_before)
@@ -1738,6 +1740,7 @@ enum_set_internal_obj_vars(class_T *en, object_T *enval)
 
 /*
  * Handle ":class" and ":abstract class" up to ":endclass".
+ * Handle ":enum" up to ":endenum".
  * Handle ":interface" up to ":endinterface".
  */
     void
@@ -2646,7 +2649,7 @@ oc_member_type_by_idx(
  * Type aliases (:type)
  */
 
-    void
+    static void
 typealias_free(typealias_T *ta)
 {
     // ta->ta_type is freed in clear_type_list()
@@ -2665,7 +2668,7 @@ typealias_unref(typealias_T *ta)
  * Handle ":type".  Create an alias for a type specification.
  */
     void
-ex_type(exarg_T *eap UNUSED)
+ex_type(exarg_T *eap)
 {
     char_u	*arg = eap->arg;
 
@@ -2777,12 +2780,13 @@ done:
  * "rettv".  If "is_object" is TRUE, then the object member variable table is
  * searched.  Otherwise the class member variable table is searched.
  */
-    static int
+    int
 get_member_tv(
     class_T	*cl,
     int		is_object,
     char_u	*name,
     size_t	namelen,
+    class_T	*current_class,
     typval_T	*rettv)
 {
     ocmember_T *m;
@@ -2793,7 +2797,8 @@ get_member_tv(
     if (m == NULL)
 	return FAIL;
 
-    if (*name == '_')
+    if (*name == '_' && (current_class == NULL ||
+				!class_instance_of(current_class, cl)))
     {
 	emsg_var_cl_define(e_cannot_access_protected_variable_str,
 							m->ocm_name, 0, cl);
@@ -2873,7 +2878,7 @@ call_oc_method(
 
     if (ocm == NULL && *fp->uf_name == '_')
     {
-	// Cannot access a private method outside of a class
+	// Cannot access a protected method outside of a class
 	semsg(_(e_cannot_access_protected_method_str), fp->uf_name);
 	return FAIL;
     }
@@ -2916,6 +2921,33 @@ call_oc_method(
     return OK;
 }
 
+/*
+ * Create a partial typval for "obj.obj_method" and store it in "rettv".
+ * Returns OK on success and FAIL on memory allocation failure.
+ */
+    int
+obj_method_to_partial_tv(object_T *obj, ufunc_T *obj_method, typval_T *rettv)
+{
+    partial_T *pt = ALLOC_CLEAR_ONE(partial_T);
+    if (pt == NULL)
+	return FAIL;
+
+    pt->pt_refcount = 1;
+    if (obj != NULL)
+    {
+	pt->pt_obj = obj;
+	++pt->pt_obj->obj_refcount;
+    }
+    pt->pt_auto = TRUE;
+    pt->pt_func = obj_method;
+    func_ptr_ref(pt->pt_func);
+
+    rettv->v_type = VAR_PARTIAL;
+    rettv->vval.v_partial = pt;
+
+    return OK;
+}
+
 /*
  * Evaluate what comes after a class:
  * - class member: SomeClass.varname
@@ -2978,7 +3010,7 @@ class_object_index(
 	// Search in the object member variable table and the class member
 	// variable table.
 	int is_object = rettv->v_type == VAR_OBJECT;
-	if (get_member_tv(cl, is_object, name, len, rettv) == OK)
+	if (get_member_tv(cl, is_object, name, len, NULL, rettv) == OK)
 	{
 	    *arg = name_end;
 	    return OK;
@@ -2989,28 +3021,17 @@ class_object_index(
 	ufunc_T	*fp = method_lookup(cl, rettv->v_type, name, len, &fidx);
 	if (fp != NULL)
 	{
-	    // Private methods are not accessible outside the class
+	    // Protected methods are not accessible outside the class
 	    if (*name == '_')
 	    {
 		semsg(_(e_cannot_access_protected_method_str), fp->uf_name);
 		return FAIL;
 	    }
 
-	    partial_T	*pt = ALLOC_CLEAR_ONE(partial_T);
-	    if (pt == NULL)
+	    if (obj_method_to_partial_tv(is_object ? rettv->vval.v_object :
+						NULL, fp, rettv) == FAIL)
 		return FAIL;
 
-	    pt->pt_refcount = 1;
-	    if (is_object)
-	    {
-		pt->pt_obj = rettv->vval.v_object;
-		++pt->pt_obj->obj_refcount;
-	    }
-	    pt->pt_auto = TRUE;
-	    pt->pt_func = fp;
-	    func_ptr_ref(pt->pt_func);
-	    rettv->v_type = VAR_PARTIAL;
-	    rettv->vval.v_partial = pt;
 	    *arg = name_end;
 	    return OK;
 	}
@@ -3101,7 +3122,7 @@ class_member_lookup(class_T *cl, char_u *name, size_t namelen, int *idx)
 	{
 	    ret_m = m;
 	    ret_idx = i;
-            break;
+	    break;
 	}
     }
     if (idx != NULL)
@@ -3184,11 +3205,11 @@ object_member_lookup(class_T *cl, char_u *name, size_t namelen, int *idx)
 	    }
 	}
 	else if (STRCMP(name, m->ocm_name) == 0)
-        {
+	{
 	    ret_m = m;
 	    ret_idx = i;
-            break;
-        }
+	    break;
+	}
     }
     if (idx != NULL)
 	*idx = ret_idx;
@@ -3295,7 +3316,7 @@ class_defining_member(class_T *cl, char_u *name, size_t len, ocmember_T **p_m)
 	    if (( m = class_member_lookup(super, name, len, NULL)) != NULL)
 	    {
 		cl_tmp = super;
-		vartype = VAR_OBJECT;
+		vartype = VAR_CLASS;
 	    }
 	}
 	if (cl_tmp == NULL)
@@ -3684,7 +3705,7 @@ method_not_found_msg(class_T *cl, vartype_T v_type, char_u *name, size_t len)
     }
     else
 	semsg(_(e_method_not_found_on_class_str_str), method_name,
-	        cl->class_name);
+		cl->class_name);
     vim_free(method_name);
 }
 
